!Full sky lensing routines by Gayoung Chon (gchon@mrao.cam.ac.uk) and AC.
!Ref: astro-ph/0001303 by W. Hu. 
!For better derivations see also astro-ph/0301064 and astro-ph/0301031

!Adapted for CAMB and optimized by AL.

!Uses the un-lensed Cls and the computed lensing potential power spectrum.
!Usual values of k_max are fine for all but the lensed BB Cls
!To get the lensed BB accurate around l=1000 you need to go to l_max >2000, and
!higher for higher l. Since this probes small scales in the lensing power spectrum you
!also need to go to higher k_max - for concordance models something like 
!k_eta_max_scalar=10000. At l>1000 you can expect to need higher k_max, and for 
!non-linear evolution to cause a significant error.

!The result is normalization dependent, but the quadratic part is simply given by
!(Cl_lensed - Cl_unlensed), so if the initial power spectrum is scaled P -> s P, then
! Cl_lensed -> s*(Cl_unlensed) + s^2*(Cl_lensed - Cl_unlensed)

!Uses f90 version of "J1-RECURSION OF 3J-COEFFICIENTS" by K. Schulten and R.G. Gordon 
!obtainable from the CPC program library (www.cpc.cs.qub.ac.uk).

module lensing
use Precision
implicit none
private
 logical  :: lensing_includes_tensors = .true.
 integer :: lmax_donelnfa = 0
 real(dl), dimension(:), allocatable  :: lnfa

public lens_Cls, lensing_includes_tensors
contains

subroutine lens_Cls
  use ModelParams
  use ModelData
  use lvalues
  use InitialPower
  integer maxl, i, in, lmin, max_lensed_ix, maxl_phi
  real(dl) , dimension (:,:,:), allocatable :: bare_cls
  real(dl) pp(CP%InitPower%nn,CP%Max_l)
  real(dl) asum(CP%InitPower%nn), RR(CP%InitPower%nn), roots(CP%Max_l)
  real(dl) asum_TE(CP%InitPower%nn), asum_EE(CP%InitPower%nn), asum_BB(CP%InitPower%nn)
  integer l1,l2,al,j, j1, k, hk, llp_1, llp_al, g1
  real(dl)  F, fct
  real(dl) g2l,g2l1, norm
  real(dl) a3j(CP%Max_l*2+1), tF, expF
  logical DoPol
  real(dl) iContribs(lSamp%l0,CP%InitPower%nn, 1:4), intcontrib(2:lSamp%l(lSamp%l0))
  real(dl) , dimension (:,:,:), allocatable :: iCl_lensed
  integer max_j_contribs

  real(sp) etime,timeprev,tarray(2)
  external etime

 
  if (DebugMsgs) timeprev=etime(tarray)

  DoPol = CP%AccuratePolarization


  maxl = CP%Max_l
 
  if (allocated(Cl_lensed)) deallocate(Cl_lensed)


  allocate(bare_cls(CP%InitPower%nn,maxl,1:4))
  
  RR = 0
  do j=2,maxl
     norm = OutputDenominator/(j*(j+1))
     if (lensing_includes_tensors .and. CP%WantTensors .and. j<= CP%Max_l_tensor) then !Use total Cls
      bare_cls(:,j,CT_Temp:CT_E) = (Cl_scalar(j,:,C_Temp:C_E) + &
           Cl_tensor(j,:,CT_Temp:CT_E))*norm
      bare_cls(:,j,CT_B) = Cl_tensor(j,:,CT_B)*norm
      bare_cls(:,j,CT_Cross) =  (Cl_scalar(j,:,C_Cross) + &
          Cl_tensor(j,:,CT_Cross))*norm
     else
      bare_cls(:,j,CT_Temp:CT_E) = Cl_scalar(j,:,C_Temp:C_E)*norm
      bare_cls(:,j,CT_B) = 0
      bare_cls(:,j,CT_Cross) =  Cl_scalar(j,:,C_Cross)*norm
     end if
     pp(:,j) = Cl_scalar(j,:,C_Phi)/real(j**2,dl)**2
     RR = RR + j*(j+1)*real(2*j+1,dl)*pp(:,j)
     roots(j) = sqrt(real(2*j+1,dl))
  end do

  RR = RR/2/fourpi
 
  if (maxl > lmax_donelnfa) then 
   !Get ln factorials
   if (allocated(lnfa)) deallocate(lnfa)
   allocate(lnfa(0:maxl*3+1))
   lmax_donelnfa = maxl 
   lnfa(0) = 0
   do i=1,CP%Max_l*3+1 
     lnfa(i)=lnfa(i-1) + log(real(i,dl))
   end do
  end if
   
  max_lensed_ix = lSamp%l0-1
  do while(lSamp%l(max_lensed_ix) > maxl -250)
     max_lensed_ix = max_lensed_ix -1
  end do
  lmax_lensed = lSamp%l(max_lensed_ix)

  allocate(iCl_lensed(max_lensed_ix, CP%InitPower%nn, 1:4))

  max_j_contribs = lSamp%l0-1
  if (.not. DoPol) then
           maxl_phi = min(maxl,nint(max(600,(maxl*2)/5)*scale*AccuracyBoost))
           do while (lSamp%l(max_j_contribs) > maxl_phi)
              max_j_contribs=max_j_contribs-1
           end do
  end if

  !$OMP PARALLEL DO DEFAULT(SHARED), SCHEDULE(DYNAMIC), SHARED(max_j_contribs) &
  !$OMP PRIVATE(al,g1,llp_al,llp_1,g2l,asum,l1,g2l1,l2,k,hk,F,fct,lmin), &
  !$OMP PRIVATE(asum_EE,asum_BB,asum_TE,expF,tF, a3j, iContribs,in,intcontrib)
  do j=max_lensed_ix,1,-1  
     !Only compute lensed spectra at lSamp%l(j). Start with slow ones.
     
     al=lSamp%l(j)

     llp_al = al*(al+1)
     g2l=sqrt((2*al+1)/fourpi)

     asum = 0
     asum_EE = 0
     asum_BB = 0
     asum_TE = 0

   
     do j1 = 1, max_j_contribs
        !  Contributions to C_al are a smooth function of l_1 - so interpolate
        l1=lSamp%l(j1)

        llp_1 = l1*(l1+1)
        g2l1=roots(l1)

        lmin = max(abs(al-l1),2)

        if (DoPol) then

          call rec3jj(a3j(lmin),dble(l1),dble(al),0._dl,2._dl)
          do l2= lmin, min(maxl,al+l1)
              g1 = llp_1+l2*(l2+1)-llp_al
              if (g1 == 0 ) cycle

              k=al+l1+l2
              fct=g1*g2l*g2l1*roots(l2)/2
              tF = fct*a3j(l2)

              if (mod(k,2)==0) then

                 hk = k/2
                 F = lnfa(hk)-lnfa(hk-al)-lnfa(hk-l1)-lnfa(hk-l2)+(lnfa(k-2*al)+lnfa(k-2*l1)&
                  & +lnfa(k-2*l2)-lnfa(k+1))/2
                 
                 expF = exp(F)

                 asum=asum + bare_cls(:,l2,C_Temp)*(expF*fct)**2
        
                 asum_EE = asum_EE + bare_cls(:,l2,CT_E)*tF**2
                 asum_BB = asum_BB + bare_cls(:,l2,CT_B)*tF**2
                 if (mod(hk,2)/=0) tF=-tF
                 asum_TE = asum_TE + bare_cls(:,l2,CT_Cross)*expF*fct*tF

              else
                
                 asum_BB = asum_BB + bare_cls(:,l2,CT_E)*tF**2
                 asum_EE = asum_EE +bare_cls(:,l2,CT_B)*tF**2 

              end if
              
          end do

        else !No polarization
          do l2= lmin +mod(al+l1+lmin,2),min(maxl,al+l1), 2 
             !Only do lSamp%l's where al + l1 + l2 is even
             
              g1 = llp_1+l2*(l2+1)-llp_al
            
              if (g1 == 0 ) cycle  !Contribution is zero

              k=al+l1+l2
              hk=k/2
           
              fct=g1*g2l*g2l1*roots(l2)/2
              expF = exp(2*(lnfa(hk)-lnfa(hk-al)-lnfa(hk-l1)-lnfa(hk-l2))+lnfa(k-2*al)+lnfa(k-2*l1)&
                  & +lnfa(k-2*l2)-lnfa(k+1))
              asum=asum + bare_cls(:,l2,CT_Temp)*expF *fct**2
          
          end do
          end if !No polarization


             iContribs(j1,:,CT_Temp) = asum*pp(:,l1)
             if (DoPol) then
                iContribs(j1,:,CT_E) = asum_EE*pp(:,l1)
                iContribs(j1,:,CT_B) = asum_BB*pp(:,l1)
                iContribs(j1,:,CT_Cross) = asum_TE*pp(:,l1)
             end if
             asum = 0
             asum_EE = 0
             asum_BB = 0
             asum_TE = 0

        
       end do

          
       !Interpolate contributions to sum and add up
         do in=1, CP%InitPower%nn
          
            call InterpolateClArr(lSamp,iContribs(1,in,CT_Temp),intcontrib,max_j_contribs)
            asum(in) = sum(intcontrib(2:lSamp%l(max_j_contribs)))
            if (DoPol) then
               call InterpolateClArr(lSamp,iContribs(1,in,CT_E),intcontrib,max_j_contribs)
               asum_EE(in) = sum(intcontrib(2:lSamp%l(max_j_contribs)))
               call InterpolateClArr(lSamp,iContribs(1,in,CT_B),intcontrib,max_j_contribs)
               asum_BB(in) = sum(intcontrib(2:lSamp%l(max_j_contribs)))
               call InterpolateClArr(lSamp,iContribs(1,in,CT_Cross),intcontrib,max_j_contribs)
               asum_TE(in) = sum(intcontrib(2:lSamp%l(max_j_contribs)))
            end if
         end do


     iCl_lensed(j,:,CT_Temp) = ((1-al*(al+1)*RR)*bare_cls(:,al,CT_Temp)  & !Linear part
              + asum/(2*al+1))*llp_al/OutputDenominator !add quadratic part and *l(l+1)/2pi
     if (DoPol) then
        iCl_lensed(j,:,CT_E) = ((1-(al**2+al-4)*RR)*bare_cls(:,al,CT_E)  & 
              + asum_EE/(2*al+1))*llp_al/OutputDenominator
        iCl_lensed(j,:,CT_B) = ((1-(al**2+al-4)*RR)*bare_cls(:,al,CT_B)  & 
              + asum_BB/(2*al+1))*llp_al/OutputDenominator
        iCl_lensed(j,:,CT_Cross) =  ((1-(al**2+al-2)*RR)*bare_cls(:,al,CT_Cross) &
                + asum_TE/(2*al+1))*llp_al/OutputDenominator
     else
        iCl_lensed(j,:,CT_E:CT_Cross) = bare_cls(:,al,CT_E:CT_Cross)
     end if

  end do
  !$OMP END PARALLEL DO

  deallocate(bare_cls)

  allocate(Cl_lensed(2:lmax_lensed,CP%InitPower%nn,1:4))

  !Interpolate to get final spectrum
  do in=1, CP%InitPower%nn
     do j = CT_Temp, CT_Cross
      call InterpolateClArr(lSamp,iCl_lensed(1,in,j),Cl_lensed(2, in, j),max_lensed_ix)
     end do
  end do

  deallocate(iCl_lensed)

  if (DebugMsgs) then
         write(*,*) etime(tarray)-timeprev,' Timing for lensing'
   end if

end subroutine lens_Cls


end module lensing


subroutine rec3jj(thrcof,l2,l3,m2,m3)
  !Recursive evaluation of 3j symbols. Does minimal error checking on input parameters.
  use Precision

  implicit real(dl) (a-h)
  implicit real(dl) (o-z)
  real(dl), intent(in) :: l2,l3, m2,m3
  real(dl) :: l1, m1, l1min,l1max, newfac,lmatch
  real(dl), dimension(*) :: thrcof
  integer ier

  data zero,eps,one /0.0d0,0.01d0,1d0/

! routine to generate set of 3j-coeffs (l1,l2,l3\\ m1,m2,m3)

! by recursion from l1min = max(abs(l2-l3),abs(m1)) 
!                to l1max = l2+l3
! the resulting 3j-coeffs are stored as thrcof(l1-l1min+1)

! to achieve the numerical stability, the recursion will proceed
! simultaneously forwards and backwards, starting from l1min and l1max
! respectively.
!
! lmatch is the l1-value at which forward and backward recursion are matched.
!
! ndim is the length of the array thrcof
!
! ier = -1 for all 3j vanish(l2-abs(m2)<0, l3-abs(m3)<0 or not integer)
! ier = -2 if possible 3j's exceed ndim
! ier >= 0 otherwise
!
  data tiny,srtiny /1.0d-10,1.0d-5/
  data huge,srhuge /1.0d10,1.0d5/

  lmatch = zero
  m1 = -(m2+m3)

! check relative magnitude of l and m values
  ier = 0
 
  if (l2 < abs(m2) .or. l3 < m3) then
  ier = -1
  return
  end if

! limits for l1
  l1min = max(abs(l2-l3),abs(m1))
  l1max = l2+l3

  if (l1min >= l1max-eps) then
   if (l1min/=l1max) then
   ier = -1
   return
   end if

! reached if l1 can take only one value, i.e.l1min=l1max
  thrcof(1) = (-1)**nint(abs(l2+m2-l3+m3))/sqrt(l1min+l2+l3+one)
  l1cmin = l1min
  l1cmax = l1max
  return

  end if

  nfin = int(l1max-l1min+one)
 

! starting forward recursion from l1min taking nstep1 steps
  l1 = l1min
  thrcof(1) = srtiny
  sum1 = (2*l1 + 1)*tiny

  lstep = 1

30  lstep = lstep+1
  l1 = l1+1

  oldfac = newfac
  a1 = (l1+l2+l3+1)*(l1-l2+l3)*(l1+l2-l3)*(-l1+l2+l3+1)
  a2 = (l1+m1)*(l1-m1)
  newfac = sqrt(a1*a2)
  if (l1 < one+eps) then
     !IF L1 = 1  (L1-1) HAS TO BE FACTORED OUT OF DV, HENCE
     c1 = -(2*l1-1)*l1*(m3-m2)/newfac
  else

   dv = -l2*(l2+1)*m1 + l3*(l3+1)*m1 + l1*(l1-1)*(m3-m2)
   denom = (l1-1)*newfac

   if (lstep > 2) c1old = abs(c1)
   c1 = -(2*l1-1)*dv/denom

  end if

  if (lstep<= 2) then

! if l1=l1min+1 the third term in the recursion eqn vanishes, hence
   x = srtiny*c1
   thrcof(2) = x
   sum1 = sum1+tiny*(2*l1+1)*c1*c1
   if(lstep==nfin) then
      sumuni=sum1
      go to 230
   end if
   goto 30

  end if

  c2 = -l1*oldfac/denom

! recursion to the next 3j-coeff x  
  x = c1*thrcof(lstep-1) + c2*thrcof(lstep-2)
  thrcof(lstep) = x
  sumfor = sum1
  sum1 = sum1 + (2*l1+1)*x*x
  if (lstep/=nfin) then

! see if last unnormalised 3j-coeff exceeds srhuge
  if (abs(x) >= srhuge) then
     
     ! REACHED IF LAST 3J-COEFFICIENT LARGER THAN SRHUGE
     ! SO THAT THE RECURSION SERIES THRCOF(1), ... , THRCOF(LSTEP)
     ! HAS TO BE RESCALED TO PREVENT OVERFLOW
     
     ier = ier+1
     do i = 1, lstep
        if (abs(thrcof(i)) < srtiny) thrcof(i)= zero
        thrcof(i) = thrcof(i)/srhuge
     end do

     sum1 = sum1/huge
     sumfor = sumfor/huge
     x = x/srhuge

  end if

! as long as abs(c1) is decreasing, the recursion proceeds towards increasing
! 3j-valuse and so is numerically stable. Once an increase of abs(c1) is 
! detected, the recursion direction is reversed.

 if (c1old > abs(c1)) goto 30

 end if !lstep/=nfin

! keep three 3j-coeffs around lmatch for comparison with backward recursion

  lmatch = l1-1
  x1 = x
  x2 = thrcof(lstep-1)
  x3 = thrcof(lstep-2)
  nstep2 = nfin-lstep+3

! --------------------------------------------------------------------------
!
! starting backward recursion from l1max taking nstep2 stpes, so that
! forward and backward recursion overlap at 3 points 
! l1 = lmatch-1, lmatch, lmatch+1

  nfinp1 = nfin+1
  nfinp2 = nfin+2
  nfinp3 = nfin+3
  l1 = l1max
  thrcof(nfin) = srtiny
  sum2 = tiny*(2*l1+1)
 
  l1 = l1+2
  lstep=1

  do
  lstep = lstep + 1
  l1= l1-1

  oldfac = newfac
  a1s = (l1+l2+l3)*(l1-l2+l3-1)*(l1+l2-l3-1)*(-l1+l2+l3+2)
  a2s = (l1+m1-1)*(l1-m1-1)
  newfac = sqrt(a1s*a2s)

  dv = -l2*(l2+1)*m1 + l3*(l3+1)*m1 +l1*(l1-1)*(m3-m2)

  denom = l1*newfac
  c1 = -(2*l1-1)*dv/denom
  if (lstep <= 2) then

     ! if l2=l2max+1, the third term in the recursion vanishes
     
     y = srtiny*c1
     thrcof(nfin-1) = y
     sumbac = sum2
     sum2 = sum2 + tiny*(2*l1-3)*c1*c1

     cycle

  end if

  c2 = -(l1-1)*oldfac/denom

! recursion to the next 3j-coeff y
  y = c1*thrcof(nfinp2-lstep)+c2*thrcof(nfinp3-lstep)

  if (lstep==nstep2) exit
  
  thrcof(nfinp1-lstep) = y
  sumbac = sum2
  sum2 = sum2+(2*l1-3)*y*y

! see if last unnormalised 3j-coeff exceeds srhuge
  if (abs(y) >= srhuge) then
     
     ! reached if 3j-coeff larger than srhuge so that the recursion series
     ! thrcof(nfin),..., thrcof(nfin-lstep+1) has to be rescaled to prevent overflow
     
     ier=ier+1
     do i = 1, lstep
        index=nfin-i+1
        if (abs(thrcof(index)) < srtiny) thrcof(index)=zero
        thrcof(index) = thrcof(index)/srhuge
     end do

     sum2=sum2/huge
     sumbac=sumbac/huge

  end if

  end do

! the forward recursion 3j-coeffs x1, x2, x3 are to be matched with the 
! corresponding backward recursion vals y1, y2, y3

  y3 = y
  y2 = thrcof(nfinp2-lstep)
  y1 = thrcof(nfinp3-lstep)

! determine now ratio such that yi=ratio*xi (i=1,2,3) holds with minimal error

  ratio = (x1*y1+x2*y2+x3*y3)/(x1*x1+x2*x2+x3*x3)
  nlim = nfin-nstep2+1

  if (abs(ratio) >= 1) then

   thrcof(1:nlim) = ratio*thrcof(1:nlim) 
   sumuni = ratio*ratio*sumfor + sumbac

  else

  nlim = nlim+1
  ratio = 1/ratio
  do n = nlim, nfin
     thrcof(n) = ratio*thrcof(n)
  end do
  sumuni = sumfor + ratio*ratio*sumbac

  end if
! normalise 3j-coeffs

230 cnorm = 1/sqrt(sumuni)

! sign convention for last 3j-coeff determines overall phase

  sign1 = sign(one,thrcof(nfin))
  sign2 = (-1)**nint(abs(l2+m2-l3+m3))
  if (sign1*sign2 <= 0) then
    cnorm = -cnorm
  end if
  if (abs(cnorm) >= one) then
     thrcof(1:nfin) = cnorm*thrcof(1:nfin)
     return
  end if

  thresh = tiny/abs(cnorm)

  do n = 1, nfin
     if (abs(thrcof(n)) < thresh) thrcof(n) = zero
     thrcof(n) = cnorm*thrcof(n)
  end do
  return 

end subroutine rec3jj



